/******************************************************************************
** Monitor.h -- a simple DAB+ Radcap monitoring program                    **
*******************************************************************************
** Copyright (C) 2009 Innes Corporation Pty Ltd                              **
**                                                                           **
** This software is provided 'as-is', without any express or implied         **
** warranty. In no event will the authors be held liable for any damages     **
** arising from the use of this software.                                    **
**                                                                           **
** Permission is granted to anyone to use this software for any purpose,     **
** including commercial applications, and to alter it and redistribute it    **
** freely, subject to the following restrictions:                            **
**   1. The origin of this software must not be misrepresented; you must not **
**      claim that you wrote the original software. If you use this software **
**      in a product, an acknowledgment in the product documentation would   **
**      be appreciated but is not required.                                  **
**   2. Altered source versions must be plainly marked as such, and must not **
**      be misrepresented as being the original software.                    **
**   3. This notice may not be removed or altered from any source            **
**      distribution.                                                        **
******************************************************************************/

#define UM_BUTTONDOWN WM_APP+6
#define UM_SLIDESHOW WM_APP+14
#define UM_OPENCHANNEL WM_APP+15
#define DLS_BOX 788
#define DLS_LINE_1 789
#define DLS_LINE_2 790
#define DLS_LINE_3 791
#define DLS_LINE_4 792
#define DLS_LINE_5 793
#define MOT_BOX 794
#define DATE_TIME 795
#define SERVICE_NAME 796
#define ENSEMBLE_ID 797
#define SERVICE_ID 798
#define SUBCHANNEL_ID 799
#define SAMPLING_RATE 800
#define BIT_RATE 801
#define FORMAT 802
#define UNCORRECTED_ERRORS 803
#define INFO_BOX 804
#define CHANNEL 805
#define ENSEMBLE_NAME 806
#define CORRELATOR 807
#define SPECTRUM 808
#define CHANNEL_BUTTON_BOX 809

const int MAX_ENSEMBLES = 4;
const UINT ServiceMax = 256;
const WCHAR *MonthName[12] = { L"Jan", L"Feb", L"Mar", L"Apr", L"May", L"Jun", L"Jul", L"Aug", L"Sep", L"Oct", L"Nov", L"Dec" };
const WCHAR *FormatStringMP2[] = { L"MP2 Monaural", L"MP2 Discrete Stereo", L"", L"" };
const WCHAR *FormatStringAAC[] = { L"AAC Monaural", L"AAC Discrete Stereo", L"AAC Parametric Stereo", L"AAC Surround" };
const WCHAR *FormatStringAACSBR[] = { L"AAC Monaural SBR", L"AAC Discrete Stereo SBR", L"AAC Parametric Stereo SBR", L"AAC Surround SBR" };

class DlsDecoder {
	unsigned FirstLast;
	unsigned SegmentLength;
	unsigned SegNum;
	unsigned ByteCount;
	unsigned Charset;
	unsigned short Crc;
	char DlsText[8][16];
	char Display[5][33];
	bool ValidSegment[8];
	unsigned SegLen[8];
	HWND DlsWindow;
public:
	DlsDecoder (HWND hwnd) : FirstLast (0), SegmentLength (0), SegNum (0), ByteCount (0), Charset (0), Crc (0xffff), DlsWindow (hwnd) {
		memset (DlsText, ' ', sizeof DlsText);
		memset (ValidSegment, 0, sizeof ValidSegment);
		memset (Display, 0, sizeof Display);
		memset (SegLen, 0, sizeof SegLen);
	}
	~DlsDecoder ();
	void Start (unsigned char *pXPad, unsigned Length);
	void Continue (unsigned char *pXPad, unsigned Length);
	void Flush () {
		memset (DlsText, ' ', sizeof DlsText);
		memset (ValidSegment, 0, sizeof ValidSegment);
		memset (Display, 0, sizeof Display);
		memset (SegLen, 0, sizeof SegLen);
		EraseDls ();
	}
private:
	void EraseDls ();
};


struct SegmentEntry {
	UCHAR *Container;
	size_t Offset;
	unsigned Size;
	SegmentEntry () : Container (NULL), Offset (0), Size (0) {}
	~SegmentEntry ();
};


class SlideDecoder {
	UCHAR DataGroupLengthIndicator[4];	// This might be spread over two fixed XPAD subfields
	unsigned DataGroupLengthIndicatorByteCount;
	unsigned DataGroupLength;
	UCHAR *CurrentDataGroup;
	unsigned CurrentDataGroupIndex;
	unsigned CurrentTransportId;
	SegmentEntry HeaderSegment;
	SegmentEntry BodySegment[32768];
	unsigned SegmentCount;
	HWND MotWindow;
public:
	SlideDecoder (HWND hwnd) : DataGroupLength (0), CurrentDataGroup (NULL), CurrentDataGroupIndex (0), MotWindow (hwnd), SegmentCount (0), CurrentTransportId (-1), DataGroupLengthIndicatorByteCount (0) {
		memset (&HeaderSegment, 0, sizeof HeaderSegment);
		memset (BodySegment, 0, sizeof BodySegment);
	}
	~SlideDecoder ();
	void Flush ();
	void Start (unsigned char *pXPad, unsigned Length, bool Continuation);
	void Continue (unsigned char *pXPad, unsigned Length);
	void DecodeDataGroup (unsigned char *DataGroup, unsigned Length);
	void ProcessObject ();
};


class PadDecoder {
protected:
	DlsDecoder Dls;
	SlideDecoder MOT;
	HWND MainWindow;
	unsigned XPadLength;
	unsigned char XPadInd;
	bool Music;
	bool Speech;
	bool CI;
	unsigned NextAppTy;
public:
	PadDecoder (HWND hwnd) : MainWindow (hwnd), XPadLength (0), XPadInd (0), Music (false), Speech (false), CI (false), NextAppTy (0), Dls (hwnd ? GetDlgItem (hwnd, DLS_BOX) : NULL), MOT (hwnd ? GetDlgItem (hwnd, MOT_BOX) : NULL) {}
	void DecodeFPad (unsigned char *FPad, unsigned DseLength);
	void DecodeXPad (unsigned char *XPad, unsigned FieldLength);
	void DecodeSubfield (unsigned char *pXPad, unsigned AppTy, unsigned Length, bool Continuation);
	void FlushPad ();
};


class GenericButton {
protected:
	HWND hParent;
	DWORD CtlID;
	bool Pressed;
	bool HasFocus;
public:
	GenericButton (HWND ParentWindow, DWORD Identifier) :
	hParent (ParentWindow), CtlID (Identifier), Pressed (false), HasFocus (false) {
	}
	virtual ~GenericButton () {}
	void Press () {
		if (!Pressed) {
			Pressed = true;
			HasFocus = true;
			InvalidateRect (GetDlgItem (hParent, CtlID), NULL, FALSE);
			SendMessage (hParent, UM_BUTTONDOWN, CtlID, 0);
		}
	}
	void Release () {
		if (Pressed) {
			Pressed = false;
			if (HasFocus) {
				HasFocus = false;
				InvalidateRect (GetDlgItem (hParent, CtlID), NULL, FALSE);
				SendMessage (hParent, WM_COMMAND, MAKELONG (CtlID, BN_CLICKED), (LPARAM) GetDlgItem (hParent, CtlID));
			}
		}
	}
	void SetFocus (bool Focus) {
		if (Pressed) {
			bool Changed = Focus != HasFocus;
			HasFocus = Focus;
			if (Changed) InvalidateRect (GetDlgItem (hParent, CtlID), NULL, FALSE);
		}
	}
	virtual void Draw (HDC hDC) = 0;
};


class ChannelButton : public GenericButton {
	HANDLE hDevice;		// The HANDLE returned by DabRadcapOpen ()
	unsigned WaveId;	// The ID used with waveInOpen ()
public:
	ChannelButton (HWND ParentWindow, DWORD Identifier, HANDLE hDev, unsigned WaveNum);
	~ChannelButton ();
	void Draw (HDC hDC);
	void ActivatePad ();
	BOOL GetDateTime (DABDateTimeInfo *pDateTime);
	BOOL GetFormat (DABFormat *pFormat);
	BOOL GetEnsembleInfoW (EnsembleIdentificationW *pInfo);
	BOOL GetServiceInfoW (ServiceInfoW *pInfo);
	BOOL GetServiceStatus (ServiceStatus *pStatus);
	UINT GetUncorrectedErrorCount ();
	unsigned GetWaveId () {
		return WaveId;
	}
	void SetWaveId (unsigned NewWaveId) {
		WaveId = NewWaveId;
	}
};


struct RSSIImage {
	HDC hMemoryDC;
	HBITMAP hBitmap;
	HBITMAP hOldBitmap;
	RSSIImage (HWND hwnd) {
		// An image of the indicator bar is stored in a memory DC and partially rendered in WM_PAINT		HDC hWindowDC = GetDC (hwnd);
		HDC hWindowDC = GetDC (hwnd);
		hMemoryDC = CreateCompatibleDC (hWindowDC);
		RECT WindowRect;
		GetClientRect (hwnd, &WindowRect);
		hBitmap = CreateCompatibleBitmap (hWindowDC, WindowRect.right - WindowRect.left, WindowRect.bottom - WindowRect.top);
		hOldBitmap = (HBITMAP) SelectObject (hMemoryDC, hBitmap);
		DeleteDC (hWindowDC);
		for (int y = 2; y < WindowRect.bottom - 2; y++) {
			if (y & 3) {
				float RedIntensity = float (WindowRect.bottom - y) / float (WindowRect.bottom - WindowRect.top);
				float GreenIntensity = 1.0f - RedIntensity * RedIntensity * RedIntensity * RedIntensity;
				for (int x = 2; x < WindowRect.right - 2; x++) {
					SetPixel (hMemoryDC, x, WindowRect.bottom - y, RGB (int (RedIntensity * 256.0f), int (GreenIntensity * 256.0f), 0));
				}
			}
		}
	}
	~RSSIImage () {
		SelectObject (hMemoryDC, hOldBitmap);
		DeleteObject (hBitmap);
		DeleteDC (hMemoryDC);
	}
};


LRESULT CALLBACK MonitorProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK CardBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK ChannelButtonUnderlayProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK RSSIIndicatorProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK SpectrumProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK DlsBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK MotBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK InfoBoxProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK GenericButtonProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
void PadThread (void *param);

const WCHAR EBULatin1[256] = {
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f,
	0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017, 0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f,
	0x0020, 0x0021, 0x0022, 0x0023, 0x00a4, 0x0025, 0x0026, 0x0027, 0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,	// Note that 0x24 ($ sign) is replaced by 0xa4 (Generic currency symbol)
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x2012, 0x005f,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x203e, 0x007f,
	0x00e1, 0x00e0, 0x00e9, 0x00e8, 0x00ed, 0x00ec, 0x00f3, 0x00f2, 0x00fa, 0x00f9, 0x00d1, 0x00c7, 0x015e, 0x00df, 0x00a1, 0x0132,
	0x00e2, 0x00e4, 0x00ea, 0x00eb, 0x00ee, 0x00ef, 0x00f4, 0x00f6, 0x00fb, 0x00fc, 0x00f1, 0x00e7, 0x015f, 0x011f, 0x0131, 0x0133,
	0x00aa, 0x03b1, 0x00a9, 0x2030, 0x011e, 0x011b, 0x0148, 0x0151, 0x03c0, 0x20ac, 0x00a3, 0x0024, 0x2190, 0x2191, 0x2192, 0x2193,
	0x00ba, 0x00b9, 0x00b2, 0x00b3, 0x00b1, 0x0130, 0x0144, 0x0171, 0x00b5, 0x00bf, 0x00f7, 0x00b0, 0x00bc, 0x00bd, 0x00be, 0x00a7,
	0x00c1, 0x00c0, 0x00c9, 0x00c8, 0x00cd, 0x00cc, 0x00d3, 0x00d2, 0x00da, 0x00d9, 0x0158, 0x010c, 0x0160, 0x017d, 0x0110, 0x013f,
	0x00c2, 0x00c4, 0x00ca, 0x00cb, 0x00ce, 0x00cf, 0x00d4, 0x00d6, 0x00db, 0x00dc, 0x0159, 0x010d, 0x0161, 0x017e, 0x0111, 0x0140,
	0x00c3, 0x00c5, 0x00c6, 0x0152, 0x0177, 0x00dd, 0x00d5, 0x00d8, 0x00de, 0x014a, 0x0154, 0x0106, 0x015a, 0x0179, 0x0166, 0x00f0,
	0x00e3, 0x00e5, 0x00e6, 0x0153, 0x0175, 0x00fd, 0x00f5, 0x00f8, 0x00fe, 0x014b, 0x0155, 0x0107, 0x015b, 0x017a, 0x0167, 0x0000
};

const WCHAR EBULatin2[256] = {
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f,
	0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017, 0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f,
	0x0020, 0x0021, 0x0022, 0x0023, 0x00a4, 0x0025, 0x0026, 0x0027, 0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,	// Note that 0x24 ($ sign) is replaced by 0xa4 (Generic currency symbol)
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x2012, 0x005f,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x203e, 0x007f,
	0x00e1, 0x00e0, 0x00e9, 0x00e8, 0x00ed, 0x00ec, 0x00f3, 0x00f2, 0x00fa, 0x00f9, 0x00d1, 0x00c7, 0x015e, 0x00df, 0x00a1, 0x0132,
	0x00e2, 0x00e4, 0x00ea, 0x00eb, 0x00ee, 0x00ef, 0x00f4, 0x00f6, 0x00fb, 0x00fc, 0x00f1, 0x00e7, 0x015f, 0x011f, 0x0131, 0x0133,
	0x00aa, 0x013e, 0x00a9, 0x2030, 0x0103, 0x011b, 0x0148, 0x0151, 0x0165, 0x20ac, 0x00a3, 0x0024, 0x2190, 0x2191, 0x2192, 0x2193,
	0x00ba, 0x00b9, 0x00b2, 0x00b3, 0x00b1, 0x0130, 0x0144, 0x0171, 0x0163, 0x00bf, 0x00f7, 0x00b0, 0x00bc, 0x00bd, 0x00be, 0x00a7,
	0x0404, 0x042f, 0x0411, 0x0427, 0x0414, 0x042d, 0x0424, 0x0403, 0x042a, 0x0418, 0x0416, 0x040c, 0x041b, 0x045b, 0x0452, 0x042b,
	0x00fd, 0x0409, 0x010f, 0x0428, 0x0426, 0x042e, 0x0429, 0x040a, 0x040f, 0x0419, 0x0417, 0x010d, 0x0161, 0x017e, 0x0111, 0x0107,
	0x03a0, 0x03b1, 0x03d0, 0x03c8, 0x03b4, 0x03b5, 0x03c6, 0x03b3, 0x03b7, 0x03b9, 0x03a3, 0x03c7, 0x03bb, 0x03bc, 0x03bd, 0x03c9,
	0x03c0, 0x03a9, 0x03c1, 0x03c3, 0x03c4, 0x03be, 0x0398, 0x0393, 0x039e, 0x03c5, 0x03b6, 0x03c2, 0x039b, 0x03a8, 0x0394, 0x0000
};

const WCHAR EBULatin3[256] = {
	0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f,
	0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017, 0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f,
	0x0020, 0x0021, 0x0022, 0x0023, 0x00a4, 0x0025, 0x0026, 0x0027, 0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,	// Note that 0x24 ($ sign) is replaced by 0xa4 (Generic currency symbol)
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x2012, 0x005f,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x203e, 0x0627,
	0x0628, 0x062a, 0x0629, 0x062b, 0x062c, 0x062d, 0x062e, 0x062f, 0x0630, 0x0631, 0x0632, 0x0633, 0x0634, 0x0635, 0x0636, 0x0637,
	0x0638, 0x0639, 0x063a, 0x0641, 0x0642, 0x0643, 0x0644, 0x0645, 0x0646, 0x0647, 0x0648, 0x064a, 0x2190, 0x2191, 0x2192, 0x2193,
	0x05d0, 0x05d1, 0x05d2, 0x05d3, 0x05d4, 0x05d5, 0x05d6, 0x05d7, 0x05d8, 0x05d9, 0x05db, 0x05da, 0x05dc, 0x05de, 0x05dd, 0x05e0,
	0x05df, 0x05e1, 0x05e2, 0x05e4, 0x05e3, 0x05e6, 0x05e5, 0x05e7, 0x05e8, 0x05e9, 0x05ea, 0x00b0, 0x00bc, 0x00bd, 0x00be, 0x00a7,
	0x0404, 0x042f, 0x0411, 0x0427, 0x0414, 0x042d, 0x0424, 0x0403, 0x042a, 0x0418, 0x0416, 0x040c, 0x041b, 0x045b, 0x0452, 0x042b,
	0x00fd, 0x0409, 0x010f, 0x0428, 0x0426, 0x042e, 0x0429, 0x040a, 0x040f, 0x0419, 0x0417, 0x010d, 0x0161, 0x017e, 0x0111, 0x0107,
	0x03a0, 0x03b1, 0x03d0, 0x03c8, 0x03b4, 0x03b5, 0x03c6, 0x03b3, 0x03b7, 0x03b9, 0x03a3, 0x03c7, 0x03bb, 0x03bc, 0x03bd, 0x03c9,
	0x03c0, 0x03a9, 0x03c1, 0x03c3, 0x03c4, 0x03be, 0x0398, 0x0393, 0x039e, 0x03c5, 0x03b6, 0x03c2, 0x039b, 0x03a8, 0x0394, 0x0000
};
